import polars as pl
import os
import io
import shutil
import tarfile 
import gzip
from logging import warning 


def create_tardb(tarpath: str) -> None:
    with tarfile.open(tarpath, mode = 'x:gz'):
        return

def list_entries(tarpath: str) -> list[str]:
    with tarfile.open(tarpath, mode = 'r:gz') as tar:
        return [tar.name for tar in tar.getmembers()]

def add_entries(tarpath: str, *files) -> None:

    # Uncompress
    temp_tar = tarpath.rstrip('.gz')
    with gzip.open(tarpath, 'rb') as f_in, open(temp_tar, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)

    members = list_entries(tarpath)
    # Add the new entry
    with tarfile.open(temp_tar, 'a') as tar:
        for file in files:
            # TODO not robust to .csv.gzip?
            if ".csv" not in file:
                warning(f"File must be a csv, tryig to upload [{file}] skipping it")
                continue 

            member_name = os.path.basename(file).rstrip(".csv")
            if member_name in members:
                warning(f"File [{file}] would be a duplicate of exisiting entry [{member_name}] use delete_entry or pick a new memeber name")
                continue

            encoded = pl.read_csv(file).write_csv().encode('utf8')
            tar_info = tarfile.TarInfo(name = member_name)
            tar_info.size = len(encoded)
            tar.addfile(tar_info, io.BytesIO(encoded))
    
    # Re-compress
    with open(temp_tar, 'rb') as f_in, gzip.open(tarpath, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)
    
    os.remove(temp_tar)

def delete_entry(tarpath: str, file):
    temp_tar_path = tarpath + '.temp'
    with tarfile.open(tarpath, 'r:gz') as tar, tarfile.open(temp_tar_path, 'w:gz') as new_tar:
        for member in tar.getmembers():
            # TODO not robust to .csv.gzip?
            if '.csv' in file:
                ignore_member =  os.path.basename(file).rstrip(".csv")
            else:
                ignore_member = file

            if member.name != ignore_member:
                new_tar.addfile(member, tar.extractfile(member))

    os.replace(temp_tar_path, tarpath)

def get_entry(tarpath: str, entry: str) -> pl.DataFrame:
    with tarfile.open(tarpath, mode = 'r:gz') as tar:
        file = tar.extractfile(entry)
        return pl.read_csv(file)



import polars as pl
import yaml
import plotly.graph_objects as go
from logging import warning
from . import tar_database
from . import plot_wrappers as pwrap

     
def parse_yaml(file_path) -> dict:
        with open(file_path, 'r') as file:
            try:
                data = yaml.safe_load(file)
                return data
            except yaml.YAMLError as e:
                print(f"Error parsing YAML file: {e}")


def parse_path(path: str) -> pl.DataFrame:
    if 'tar.gz' in path:
        source, entry = path.split(":")
        return tar_database.get_entry(source, entry)
    elif '.csv' in path:
        return pl.read_csv(path)
    else:
        warning(f"Unable to parse path type {path}")

        
def parse_template(template: dict, python_dfs: dict[pl.DataFrame] = None) -> list[go.Figure]:

    for key in ["data", "plots"]:
        if key not in template:
            raise ValueError(f"Template must contain the key [{key}]")

    if 'plots' not in template:
        raise ValueError("Template must contain the key [data]")

    # Compile data
    df_list = []
    for source, info in template['data'].items():

        if "type" not in info:
            warning(f"Data source [{source}] does not contain type information skipping")
            continue
        
        if info["type"] == 'path':
            df = parse_path(source)

        elif info["type"] == 'python':
            if python_dfs is None:
                warning(f'Data source [{source}] wants to use type python, but python_dfs is None, skipping')
                continue

            df = python_dfs.get(source)
            if df is None:
                warning(f'Could not find Data source [{source}] in python_dfs, skipping')
                continue
        else:
            warning(f'Unregonzied Data source [{source}] skipping')
            continue

        df_list.append(dict(
            df = df,
            source = source,
            alt_color = info.get('alt_color')
        ))

    # Get the plot configs going
    fig_list = [] 
    plot: dict
    for plot_name, plot in template['plots'].items():

        fig = go.Figure()

        for key in ["type", "x"]:
            if key not in plot:
                raise ValueError(f"Plot [{plot_name}] must contain key [{key}]")

        type = plot["type"]
        x = plot['x']

        _color_list = [
            '#2ca02c',  # cooked asparagus green
            '#1f77b4',  # muted blue
            '#d62728',  # brick red
            '#ff7f0e',  # safety orange
            '#9467bd',  # muted purple
            '#bcbd22',  # curry yellow-green
            '#8c564b',  # chestnut brown
            '#e377c2',  # raspberry yogurt pink
            '#7f7f7f',  # middle gray
            '#17becf'   # blue-teal
        ]
        color_list = lambda idx: _color_list[idx % len(_color_list)]
        curr_color = 0


        plot_axis = [_str for _str in plot if _str in ['y1', 'y2', 'y3', 'y4']]

        for axis in plot_axis:
            i = int(axis[1])    
            for y_ch in plot[axis]:

                # Hate this handling
                y_ch_content = None
                if isinstance(y_ch, dict):
                    y_ch_content = y_ch[next(iter(y_ch))]
                    y_ch = next(iter(y_ch))

                for info in df_list:

                    df = info['df']
                    source = info['source'] 

                    _color =  color_list(curr_color)
                    color = plot.get('color', _color)

                    if y_ch_content is not None:
                        color = y_ch_content.get(f'alt_color_{info["alt_color"]}', _color)

                    if ('{' and '}' not in y_ch) and (y_ch not in df):
                        warning(f'Unable to find [{y_ch}] or one of {"{ and }"} are missing in [{source}], skipping in the plot')
                        continue

                    if not plot.get('toggle_group', False):
                        fig.update_layout(legend=dict(groupclick="toggleitem"))

                    if type.lower() == 'lines' or 'markers':
                        pwrap.graph_scatter_by_key(
                            df = df,
                            x = x,
                            y = y_ch,
                            x_title = plot.get('x_title'),
                            y_title = plot.get('y_title'),
                            title = plot_name,
                            color = color,
                            group_name = y_ch,
                            fig = fig,
                            mode = type,
                            alt_y_name = f'{source}:{y_ch}',
                            options = plot.get('options', {}),
                            axis = i,
                        )
                curr_color += 1
        fig_list.append(fig)

    return fig_list


from __future__ import annotations
import polars as pl
import plotly.graph_objects as go
from copy import deepcopy
import re

# Exposing it so anything that uses eval can convert
from .units import convert

def parse_keys(full_str: str, ):

    if ('{' and '}' not in full_str):
        return full_str

    keys =  re.findall(r"\{(.*?)\}", full_str)
    new_str = deepcopy(full_str)
    for key in keys:
        new_str = new_str.replace('{' + key + '}', f'pl.col("{key}")')
    
    return 'df.with_columns([( + ' + new_str + ').alias("' + full_str +'")])'


def graph_scatter_by_key(
    df: pl.DataFrame,
    x: str,
    y: str,
    x_title = None,
    y_title = '',
    title = '',
    color = None,
    mode = 'lines',
    group_name: str = None,
    options: dict = None,
    fig: go.Figure = None,
    axis = 1,
    theme = 'plotly_dark',
    alt_y_name = None,
):

    if options is None:
        options = {}

    if x_title is None:
        x_title = x

    if fig is None:
        fig = go.Figure()
    
    if alt_y_name is None:
        alt_y_name = y

    y_axis_info = dict(
        title=dict(text=y_title),
        anchor="free",
        overlaying="y",
        autoshift=True,
        side="left"
    )

    if color is not None:
        y_axis_info['color'] = color

    if y not in df:
        df_evaluated = eval(parse_keys(y))
    else:
        df_evaluated = df 

    # TODO: Fix these copies? 
    data = dict(
        x = df_evaluated[x].to_list(), 
        y = df_evaluated[y].to_list(),
        name = alt_y_name,
        mode = mode,
        legendgroup = group_name,
        legendgrouptitle_text = group_name,
        **options
    )

    if mode == 'lines':
        data['line'] = dict(color = color)
    elif mode == 'markers':
        data['marker'] = dict(color = color)
    elif mode == 'lines+markers':
        data['line'] = dict(color = color)
        data['marker'] = dict(color = color)

    if axis == 1:
        fig.add_trace(go.Scatter(
            **data
        ))
        fig.update_layout(yaxis=dict(title=dict(text=y_title)))
    elif axis == 2:
        fig.add_trace(go.Scatter(
            yaxis='y2',
            **data
        ))
        fig.update_layout(
            yaxis2=y_axis_info,
        )
    elif axis == 3:
        fig.add_trace(go.Scatter(
            yaxis='y3',
            **data
        ))
        fig.update_layout(
            yaxis3=y_axis_info,
        )
    elif axis == 4:
        fig.add_trace(go.Scatter(
            yaxis='y4',
            **data
        ))
        fig.update_layout(
            yaxis4=y_axis_info,
        )

    fig.update_xaxes(title = x_title)
    fig.update_layout(
        title = title,
        template = theme,
        showlegend = True,
        hovermode='x unified',
    )

    return fig


def graph_scatter_by_keylist(
    df: pl.DataFrame,
    x: str,
    y_list: list[str],
    x_title = None,
    y_title = '',
    title = '',
    color = None,
    mode = 'lines',
    group_name: str = None,
    options: dict = None,
    fig: go.Figure = None,
    axis = 1,
    theme = 'plotly_dark',
    alt_y_dict = None,
):
    if fig is None:
        fig = go.Figure()

    if alt_y_dict is None:
        alt_y_dict = {}

    for y in y_list:
        fig = graph_scatter_by_key(
            df = df,
            x = x,
            y = y,
            x_title = x_title,
            y_title = y_title,
            title = title,
            color = color,
            mode = mode,
            group_name = group_name,
            options = options,
            fig = fig,
            axis = axis,
            theme = theme,
            alt_y_name = alt_y_dict.get(y),
        )

    return fig


def graph_all(
    df: pl.DataFrame,
    x: str,
    x_title: str = None,
    y_title: str = '',
    title: str = '',
    color: str = None,
    mode: str = 'lines',
    group_name: str = None,
    options: dict = None,
    fig: go.Figure = None,
    axis: int = 1,
    theme: str = 'plotly_dark',
    alt_y_dict: dict = None,
):
    key_list = [key.name for key in df if key.name != x]

    return graph_scatter_by_keylist(
        df = df,
        x = x,
        y_list = key_list,
        x_title = x_title,
        y_title = y_title,
        title = title,
        color = color,
        mode = mode,
        group_name = group_name,
        options = options,
        fig = fig,
        axis = axis,
        theme = theme,
        alt_y_dict = alt_y_dict,
    )

def graphs_containing(
    df, 
    x, 
    conatains_list, 
    mode: str = 'lines', 
    theme: str = 'plotly_dark',
    options: dict = None
) -> list[go.Figure]:

    figs = []
    for contents in conatains_list:
        figs.append(graph_scatter_by_keylist(
            df = df,
            x = x,
            y_list = [key.name for key in df if contents in key.name],
            mode = mode,
            options = options,
            theme = theme,
        ))
    
    return figs
    
def figs_to_html(*figs, prefix: str = None, export_path: str = None, include_plotlyjs: bool = False):

    if prefix is None:
        prefix = ''

    for fig in figs:
        prefix += fig.to_html(full_html = False, include_plotlyjs = include_plotlyjs)

    if export_path is not None:
        with open(export_path, '+w') as f:
            f.write(prefix) 
    
    return prefix
